;;; mm.el --- specialized comint.el for MM interpreters

;; Copyright (C) 1998, 1999, 2000, 2001  Free Software Foundation, Inc.

;; Author: Alex Schroeder <alex@gnu.org>
;; Maintainer: Alex Schroeder <alex@gnu.org>
;; Version: 1.6.5
;; Keywords: comm languages processes
;; URL: http://www.emacswiki.org/cgi-bin/wiki.pl?MmMode

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; Please send bug reports and bug fixes to the mailing list at
;; mm.el@gnu.org.  If you want to subscribe to the mailing list, send
;; mail to mm.el-request@gnu.org with `subscribe mm.el FIRSTNAME
;; LASTNAME' in the mail body.

;; This file provides a mm-mode and a mm-interactive-mode.  My goals
;; were two simple modes providing syntactic hilighting.  The
;; interactive mode had to provide a command-line history; the other
;; mode had to provide "send region/buffer to MM interpreter"
;; functions.  "simple" in this context means easy to use, easy to
;; maintain and little or no bells and whistles.

;; If anybody feels like extending this mm mode, take a look at the
;; above mentioned modes and write a mmx-mode on top of this one.  If
;; this proves to be difficult, please suggest changes that will
;; facilitate your plans.

;; mm-interactive-mode is used to interact with a MM interpreter
;; process in a MMi buffer (usually called `*MM*').  The MMi buffer
;; is created by calling a MM interpreter-specific entry function.  Do
;; *not* call mm-interactive-mode by itself.

;; The list of currently supported interpreters and the corresponding
;; entry function used to create the MMi buffers is shown with
;; `mm-help' (M-x mm-help).

;; Since mm-interactive-mode is built on top of the general
;; command-interpreter-in-a-buffer mode (comint mode), it shares a
;; common base functionality, and a common set of bindings, with all
;; modes derived from comint mode.  This makes these modes easier to
;; use.

;; mm-mode can be used to keep editing MM statements.  The MM
;; statements can be sent to the MM process in the MMi buffer.

;; For documentation on the functionality provided by comint mode, and
;; the hooks available for customizing it, see the file `comint.el'.

;; Hint for newbies: take a look at `dabbrev-expand', `abbrev-mode', and
;; `imenu-add-menubar-index'.

;;; Requirements for Emacs 19.34:

;; If you are using Emacs 19.34, you will have to get and install
;; the file regexp-opt.el
;; <URL:ftp://ftp.ifi.uio.no/pub/emacs/emacs-20.3/lisp/emacs-lisp/regexp-opt.el>
;; and the custom package
;; <URL:http://www.dina.kvl.dk/~abraham/custom/>.

;;; Bugs:

;; Using mm-ms (imm by Microsoft): When commands with syntax errors
;; or execution errors are executed, there is no server feedback.
;; This happens in stored procedures for example.  The server messages
;; only appear after the process is exited.  This makes things
;; somewhat unreliable.

;; ChangeLog available on request.

;;; To Do:

;; Add better hilight support for other brands; there is a bias towards
;; Oracle because that's what I use at work.  Anybody else just send in
;; your lists of reserved words, keywords and builtin functions!  As
;; long as I don't receive any feedback, everything is hilighted with
;; ANSI keywords only.  I received the list of ANSI keywords from a
;; user; if you know of any changes, let me know.

;; Add different hilighting levels.

;;; Thanks to all the people who helped me out:

;; Kai Blauberg <kai.blauberg@metla.fi>
;; <ibalaban@dalet.com>
;; Yair Friedman <yfriedma@JohnBryce.Co.Il>
;; Gregor Zych <zych@pool.informatik.rwth-aachen.de>
;; nino <nino@inform.dk>
;; Berend de Boer <berend@pobox.com>



;;; Code:

(require 'comint)
;; Need the following to allow GNU Emacs 19 to compile the file.
(require 'regexp-opt)
(require 'custom)

;;; Allow customization

(defgroup MM nil
  "Running a MM interpreter from within Emacs buffers"
  :version "20.4"
  :group 'processes)

;; These three variables will be used as defaults, if set.

(defcustom mm-user ""
  "*Default username."
  :type 'string
  :group 'MM)

(defcustom mm-password ""
  "*Default password.

Storing your password in a textfile such as ~/.emacs could be dangerous.
Customizing your password will store it in your ~/.emacs file."
  :type 'string
  :group 'MM)

(defcustom mm-database ""
  "*Default database."
  :type 'string
  :group 'MM)

(defcustom mm-server ""
  "*Default server or host."
  :type 'string
  :group 'MM)

;; misc customization of mm.el behaviour

(defcustom mm-electric-stuff nil
  "Treat some input as electric.
If set to the symbol `semicolon', then hitting `;' will send current
input in the MMi buffer to the process.
If set to the symbol `go', then hitting `go' on a line by itself will
send current input in the MMi buffer to the process.
If set to nil, then you must use \\[comint-send-input] in order to send
current input in the MMi buffer to the process."
  :type '(choice (const :tag "Nothing" nil)
		 (const :tag "The semikolon `;'" semicolon)
		 (const :tag "The string `go' by itself" go))
  :version "20.8"
  :group 'MM)

(defcustom mm-pop-to-buffer-after-send-region nil
  "*If t, pop to the buffer MM statements are sent to.

After a call to `mm-send-region' or `mm-send-buffer',
the window is split and the MMi buffer is shown.  If this
variable is not nil, that buffer's window will be selected
by calling `pop-to-buffer'.  If this variable is nil, that
buffer is shown using `display-buffer'."
  :type 'boolean
  :group 'MM)

;; imenu support for mm-mode.

(defvar mm-imenu-generic-expression
  '(("Tables" "^\\s-*create\\s-+table\\s-+\\(\\w+\\)" 1)
    ("Indexes" "^\\s-*create\\s-+index\\s-+\\(\\w+\\)" 1))
  "Define interesting points in the MM buffer for `imenu'.

This is used to set `imenu-generic-expression' when MM mode is
entered.  Subsequent changes to mm-imenu-generic-expression will not
affect existing MM buffers because imenu-generic-expression is a
local variable.")

;; history file

(defcustom mm-input-ring-file-name nil
  "*If non-nil, name of the file to read/write input history.

You have to set this variable if you want the history of your commands
saved from one Emacs session to the next.  If this variable is set,
exiting the MM interpreter in an MMi buffer will write the input
history to the specified file.  Starting a new process in a MMi buffer
will read the input history from the specified file.

This is used to initialize `comint-input-ring-file-name'.

Note that the size of the input history is determined by the variable
`comint-input-ring-size'."
  :type '(choice (const :tag "none" nil)
		 (file))
  :group 'MM)

(defcustom mm-input-ring-separator "\n--\n"
  "*Separator between commands in the history file.

If set to \"\\n\", each line in the history file will be interpreted as
one command.  Multi-line commands are split into several commands when
the input ring is initialized from a history file.

This variable used to initialize `comint-input-ring-separator'.
`comint-input-ring-separator' is part of Emacs 21; if your Emacs
does not have it, setting `mm-input-ring-separator' will have no
effect.  In that case multiline commands will be split into several
commands when the input history is read, as if you had set
`mm-input-ring-separator' to \"\\n\"."
  :type 'string
  :group 'MM)

;; The usual hooks

(defcustom mm-interactive-mode-hook '()
  "*Hook for customizing `mm-interactive-mode'."
  :type 'hook
  :group 'MM)

(defcustom mm-mode-hook '()
  "*Hook for customizing `mm-mode'."
  :type 'hook
  :group 'MM)

(defcustom mm-set-mmi-hook '()
  "*Hook for reacting to changes of `mm-buffer'.

This is called by `mm-set-mmi-buffer' when the value of `mm-buffer'
is changed."
  :type 'hook
  :group 'MM)

;; Customization for Oracle

(defcustom mm-oracle-program "mmplus"
  "*Command to start mmplus by Oracle.

Starts `mm-interactive-mode' after doing some setup.

Under NT, \"mmplus\" usually starts the mmplus \"GUI\".  In order to
start the mmplus console, use \"plus33\" or something similar.  You
will find the file in your Orant\\bin directory.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

(defcustom mm-oracle-options nil
  "*List of additional options for `mm-oracle-program'."
  :type '(repeat string)
  :version "20.8"
  :group 'MM)

;; Customization for MyMm

(defcustom mm-mymm-program "mymm"
  "*Command to start mymm by TcX.

Starts `mm-interactive-mode' after doing some setup.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

(defcustom mm-mymm-options nil
  "*List of additional options for `mm-mymm-program'.
The following list of options is reported to make things work
on Windows: \"-C\" \"-t\" \"-f\" \"-n\"."
  :type '(repeat string)
  :version "20.8"
  :group 'MM)

;; Customization for Solid

(defcustom mm-solid-program "solmm"
  "*Command to start SOLID MM Editor.

Starts `mm-interactive-mode' after doing some setup.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

;; Customization for SyBase

(defcustom mm-sybase-program "imm"
  "*Command to start imm by SyBase.

Starts `mm-interactive-mode' after doing some setup.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

(defcustom mm-sybase-options nil
  "*List of additional options for `mm-sybase-program'.
Some versions of imm might require the -n option in order to work."
  :type '(repeat string)
  :version "20.8"
  :group 'MM)

;; Customization for Informix

(defcustom mm-informix-program "dbaccess"
  "*Command to start dbaccess by Informix.

Starts `mm-interactive-mode' after doing some setup.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

;; Customization for Ingres

(defcustom mm-ingres-program "mm"
  "*Command to start mm by Ingres.

Starts `mm-interactive-mode' after doing some setup.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

;; Customization for Microsoft

(defcustom mm-ms-program "imm"
  "*Command to start imm by Microsoft.

Starts `mm-interactive-mode' after doing some setup.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

;; Customization for Postgres

(defcustom mm-postgres-program "pmm"
  "Command to start pmm by Postgres.

Starts `mm-interactive-mode' after doing some setup.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

(defcustom mm-postgres-options '("-P" "pager=off")
  "*List of additional options for `mm-postgres-program'.
The default setting includes the -P option which breaks older versions
of the pmm client (such as version 6.5.3).  The -P option is equivalent
to the --pset option.  If you want the pmm to prompt you for a user
name, add the string \"-u\" to the list of options.  If you want to
provide a user name on the command line (newer versions such as 7.1),
add your name with a \"-U\" prefix (such as \"-Umark\") to the list."
  :type '(repeat string)
  :version "20.8"
  :group 'MM)

;; Customization for Interbase

(defcustom mm-interbase-program "imm"
  "*Command to start imm by Interbase.

Starts `mm-interactive-mode' after doing some setup.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

(defcustom mm-interbase-options nil
  "*List of additional options for `mm-interbase-program'."
  :type '(repeat string)
  :version "20.8"
  :group 'MM)

;; Customization for DB2

(defcustom mm-db2-program "db2"
  "*Command to start db2 by IBM.

Starts `mm-interactive-mode' after doing some setup.

The program can also specify a TCP connection.  See `make-comint'."
  :type 'file
  :group 'MM)

(defcustom mm-db2-options nil
  "*List of additional options for `mm-db2-program'."
  :type '(repeat string)
  :version "20.8"
  :group 'MM)



;;; Variables which do not need customization

(defvar mm-user-history nil
  "History of usernames used.")

(defvar mm-database-history nil
  "History of databases used.")

(defvar mm-server-history nil
  "History of servers used.")

;; Passwords are not kept in a history.

(defvar mm-buffer nil
  "Current MMi buffer.

The global value of mm-buffer is the name of the latest MMi buffer
created.  Any MM buffer created will make a local copy of this value.
See `mm-interactive-mode' for more on multiple sessions.  If you want
to change the MMi buffer a MM mode sends its MM strings to, change
the local value of `mm-buffer' using \\[mm-set-mmi-buffer].")

(defvar mm-prompt-regexp nil
  "Prompt used to initialize `comint-prompt-regexp'.

You can change `comint-prompt-regexp' on `mm-interactive-mode-hook'.")

(defvar mm-prompt-length 0
  "Prompt used to set `left-margin' in `mm-interactive-mode'.

You can change it on `mm-interactive-mode-hook'.")

(defvar mm-alternate-buffer-name nil
  "Buffer-local string used to possibly rename the MMi buffer.

Used by `mm-rename-buffer'.")

;; Keymap for mm-interactive-mode.

(defvar mm-interactive-mode-map 
  (let ((map (make-sparse-keymap)))
    (if (functionp 'set-keymap-parent)
	(set-keymap-parent map comint-mode-map); Emacs
      (set-keymap-parents map (list comint-mode-map))); XEmacs
    (if (functionp 'set-keymap-name)
	(set-keymap-name map 'mm-interactive-mode-map)); XEmacs
    (define-key map (kbd "C-j") 'mm-accumulate-and-indent)
    (define-key map (kbd "C-c C-w") 'mm-copy-column)
    (define-key map (kbd "O") 'mm-magic-go)
    (define-key map (kbd "o") 'mm-magic-go)
    (define-key map (kbd ";") 'mm-magic-semicolon)
    map)
  "Mode map used for `mm-interactive-mode'.
Based on `comint-mode-map'.")

;; Keymap for mm-mode.

(defvar mm-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map (kbd "C-c C-c") 'mm-send-paragraph)
    (define-key map (kbd "C-c C-r") 'mm-send-region)
    (define-key map (kbd "C-c C-b") 'mm-send-buffer)
    map)
  "Mode map used for `mm-mode'.")

;; easy menu for mm-mode.

(easy-menu-define
 mm-mode-menu mm-mode-map
 "Menu for `mm-mode'."
 '("MM"
   ["Send Paragraph" mm-send-paragraph (and (buffer-live-p mm-buffer)
					     (get-buffer-process mm-buffer))]
   ["Send Region" mm-send-region (and (or (and (boundp 'mark-active); Emacs
						mark-active)
					   (mark)); XEmacs
				       (buffer-live-p mm-buffer)
				       (get-buffer-process mm-buffer))]
   ["Send Buffer" mm-send-buffer (and (buffer-live-p mm-buffer)
				       (get-buffer-process mm-buffer))]
   ["Show MMi buffer" mm-show-mmi-buffer t]
   ["Set MMi buffer" mm-set-mmi-buffer t]
   ["Pop to MMi buffer after send"
    mm-toggle-pop-to-buffer-after-send-region
    :style toggle
    :selected mm-pop-to-buffer-after-send-region]
   ("Highlighting"
    ["ANSI MM keywords" mm-highlight-ansi-keywords t]
    ["Oracle keywords" mm-highlight-oracle-keywords t]
    ["Postgres keywords" mm-highlight-postgres-keywords t])))

;; easy menu for mm-interactive-mode.

(easy-menu-define
 mm-interactive-mode-menu mm-interactive-mode-map
 "Menu for `mm-interactive-mode'."
 '("MM"
   ["Rename Buffer" mm-rename-buffer t]))

;; Abbreviations -- if you want more of them, define them in your
;; ~/.emacs file.  Abbrevs have to be enabled in your ~/.emacs, too.

(defvar mm-mode-abbrev-table nil
  "Abbrev table used in `mm-mode' and `mm-interactive-mode'.")
(if mm-mode-abbrev-table
    ()
  (let ((wrapper))
    (define-abbrev-table 'mm-mode-abbrev-table ())
    (define-abbrev mm-mode-abbrev-table  "ins" "insert" nil)
    (define-abbrev mm-mode-abbrev-table  "upd" "update" nil)
    (define-abbrev mm-mode-abbrev-table  "del" "delete" nil)
    (define-abbrev mm-mode-abbrev-table  "sel" "select" nil)))

;; Syntax Table

(defvar mm-mode-syntax-table
  (let ((table (make-syntax-table)))
    ;; C-style comments /**/ (see elisp manual "Syntax Flags"))

  (cond
   ;; XEmacs 19 & 20
   ((memq '8-bit c-emacs-features)
    (modify-syntax-entry ?/  ". 1456" table)
    (modify-syntax-entry ?*  ". 23"   table))
   ;; Emacs 19 & 20
   ((memq '1-bit c-emacs-features)
    (modify-syntax-entry ?/  ". 124b" table)
    (modify-syntax-entry ?*  ". 23"   table))
   ;; incompatible
   (t (error "MM Mode is incompatible with this version of Emacs"))
   )
    ;;(modify-syntax-entry ?/ ". 1456" table)
    ;;(modify-syntax-entry ?* ". 23" table)
    ;; double-dash starts comment
    ;;(if (string-match "XEmacs\\|Lucid" emacs-version)
    ;;	(modify-syntax-entry ?- ". 56" table)
    ;;  (modify-syntax-entry ?- ". 12b" table))
    ;; newline and formfeed end coments
    (modify-syntax-entry ?\n "> b" table)
    (modify-syntax-entry ?\f "> b" table)
    ;; single quotes (') quotes delimit strings
    (modify-syntax-entry ?' "." table)
    ;; backslash is no escape character
    (modify-syntax-entry ?\\ "." table)
    table)
  "Syntax table used in `mm-mode' and `mm-interactive-mode'.")

;; Font lock support

(defvar mm-mode-ansi-font-lock-keywords nil
  "ANSI MM keywords used by font-lock.

This variable is used by `mm-mode' and `mm-interactive-mode'.  The
regular expressions are created during compilation by calling the
function `regexp-opt'.  Therefore, take a look at the source before
you define your own mm-mode-ansi-font-lock-keywords.  You may want to
add functions and PL/MM keywords.")
(if mm-mode-ansi-font-lock-keywords
    ()
  (let ((ansi-keywords (eval-when-compile
			 (concat "\\b"
				 (regexp-opt '(
"Match" "Diff" "Extract" "ExtractM" "Traverse" "SelectLossless" "Copy" "Delete" "SelectMin"
"NGramMatch" "Union" "Subtract" "Intersect" "Compose" "Merge" "All" "Id" "Invert"
"DeleteHard" "DeleteSoft" "Select" "Subgraph" "FilterBest" "Submodel"
"Domain" "Range" "RestrictRange" "RestrictDomain" "Project" "Restrict"
"Dependencies" "Reachable" "TransitiveClosure"
"section" "some" "mmcode" "mmerror" "sum" "work") t) "\\b")))
	(ansi-reserved-words (eval-when-compile
			       (concat "\\b"
				       (regexp-opt '(
"if" "else" "for" "while" "do"
"true" "false" "include" "operator" "alias"
"dbread" "dbwrite" "read" "write" "popup" "-" "*"
"SQLDDL2Obj" "Obj2SQLDDL" "Obj2Text"
"XSD2Obj" "Obj2XSD"
"EditMap" "ShowMap" "EditSelector" "ShowSelector"
"dbconfig" "registerLanguage"

) t) "\\b")))
	(ansi-types (eval-when-compile
		      (concat "\\b"
			      (regexp-opt '(
;; ANSI Keywords that look like types
;;"character" "cursor" "dec" "int" "real"
;; ANSI Reserved Word that look like types
"char" "integer" "smallint") t) "\\b"))))
    (setq mm-mode-ansi-font-lock-keywords
	  (list (cons ansi-keywords 'font-lock-function-name-face)
		(cons ansi-reserved-words 'font-lock-keyword-face)
		(cons ansi-types 'font-lock-type-face)))))

(defvar mm-mode-oracle-font-lock-keywords nil
  "Oracle MM keywords used by font-lock.

This variable is used by `mm-mode' and `mm-interactive-mode'.  The
regular expressions are created during compilation by calling the
function `regexp-opt'.  Therefore, take a look at the source before
you define your own mm-mode-oracle-font-lock-keywords.  You may want
to add functions and PL/MM keywords.")
(if mm-mode-oracle-font-lock-keywords
    ()
  (let ((oracle-keywords (eval-when-compile
			   (concat "\\b"
				   (regexp-opt '(
"admin" "after" "allocate" "analyze" "archive" "archivelog" "backup"
"become" "before" "block" "body" "cache" "cancel" "cascade" "change"
"checkpoint" "compile" "constraint" "constraints" "contents"
"controlfile" "cycle" "database" "datafile" "dba" "disable" "dismount"
"dump" "each" "else" "elsif" "enable" "events" "except" "exceptions"
"execute" "exit" "explain" "extent" "externally" "false" "flush" "force"
"freelist" "freelists" "function" "groups" "if" "including" "initrans"
"instance" "layer" "link" "lists" "logfile" "loop" "manage" "manual"
"maxdatafiles" "maxinistances" "maxlogfiles" "maxloghistory"
"maxlogmembers" "maxtrans" "maxvalue" "minextents" "minvalue" "mount"
"new" "next" "noarchivelog" "nocache" "nocycle" "nomaxvalue"
"nominvalue" "none" "noorder" "noresetlogs" "normal" "nosort" "off"
"old" "only" "optimal" "others" "out" "own" "package" "parallel"
"pctincrease" "pctused" "plan" "pragma" "private" "profile" "quota"
"raise" "read" "recover" "referencing" "resetlogs" "restrict_references"
"restricted" "return" "returning" "reuse" "rnds" "rnps" "role" "roles"
"savepoint" "scn" "segment" "sequence" "shared" "snapshot" "sort"
"statement_id" "statistics" "stop" "storage" "subtype" "switch" "system"
"tables" "tablespace" "temporary" "thread" "time" "tracing"
"transaction" "triggers" "true" "truncate" "type" "under" "unlimited"
"until" "use" "using" "when" "while" "wnds" "wnps" "write") t) "\\b")))
	(oracle-warning-words (eval-when-compile
				 (concat "\\b"
					 (regexp-opt '(
"cursor_already_open" "dup_val_on_index" "exception" "invalid_cursor"
"invalid_number" "login_denied" "no_data_found" "not_logged_on"
"notfound" "others" "pragma" "program_error" "storage_error"
"timeout_on_resource" "too_many_rows" "transaction_backed_out"
"value_error" "zero_divide") t) "\\b")))
	(oracle-reserved-words (eval-when-compile
				 (concat "\\b"
					 (regexp-opt '(
"access" "add" "alter" "audit" "cluster" "column" "comment" "compress"
"connect" "drop" "else" "exclusive" "file" "grant"
"identified" "immediate" "increment" "index" "initial" "intersect"
"level" "lock" "long" "maxextents" "minus" "mode" "modify" "noaudit"
"nocompress" "nowait" "number" "offline" "online" "pctfree" "prior"
"raw" "rename" "resource" "revoke" "row" "rowlabel" "rownum"
"rows" "session" "share" "size" "start" "successful" "synonym" "sysdate"
"then" "trigger" "uid" "validate" "whenever") t) "\\b")))
	(oracle-types (eval-when-compile
			(concat "\\b"
				(regexp-opt '(
;; Oracle Keywords that look like types
;; Oracle Reserved Words that look like types
"binary_integer" "blob" "boolean" "constant" "date" "decimal" "rowid"
"varchar" "varchar2") t) "\\b")))
	(oracle-builtin-functions (eval-when-compile
			(concat "\\b"
				(regexp-opt '(
;; Misc Oracle builtin functions
"abs" "add_months" "ascii" "avg" "ceil" "chartorowid" "chr" "concat"
"convert" "cos" "cosh" "count" "currval" "decode" "dump" "exp" "floor"
"glb" "greatest" "greatest_lb" "hextoraw" "initcap" "instr" "instrb"
"last_day" "least" "least_ub" "length" "lengthb" "ln" "log" "lower"
"lpad" "ltrim" "lub" "max" "min" "mod" "months_between" "new_time"
"next_day" "nextval" "nls_initcap" "nls_lower" "nls_upper" "nlssort"
"nvl" "power" "rawtohex" "replace" "round" "rowidtochar" "rpad"
"rtrim" "sign" "sin" "sinh" "soundex" "mmcode" "mmerrm" "sqrt"
"stddev" "sum" "substr" "substrb" "tan" "tanh" "to_char"
"to_date" "to_label" "to_multi_byte" "to_number" "to_single_byte"
"translate" "trim" "trunc" "uid" "upper" "userenv" "variance" "vsize") t) "\\b"))))
    (setq mm-mode-oracle-font-lock-keywords
	  (append mm-mode-ansi-font-lock-keywords
		  (list (cons oracle-keywords 'font-lock-function-name-face)
			(cons oracle-warning-words 'font-lock-warning-face)
			(cons oracle-reserved-words 'font-lock-keyword-face)
			;; XEmacs doesn't have font-lock-builtin-face
			(if (string-match "XEmacs\\|Lucid" emacs-version)
			    (cons oracle-builtin-functions 'font-lock-preprocessor-face)
			  ;; GNU Emacs 19 doesn't have it either
			  (if (string-match "GNU Emacs 19" emacs-version)
			      (cons oracle-builtin-functions 'font-lock-function-name-face)
			    ;; Emacs
			    (cons oracle-builtin-functions 'font-lock-builtin-face)))
			(cons oracle-types 'font-lock-type-face))))))

(defvar mm-mode-postgres-font-lock-keywords nil
  "Postgres MM keywords used by font-lock.

This variable is used by `mm-mode' and `mm-interactive-mode'.  The
regular expressions are created during compilation by calling the
function `regexp-opt'.  Therefore, take a look at the source before
you define your own mm-mode-postgres-font-lock-keywords.")

(if mm-mode-postgres-font-lock-keywords
    ()
  (let ((postgres-reserved-words (eval-when-compile
				 (concat "\\b"
					 (regexp-opt '(
"language"
) t) "\\b")))
	(postgres-types (eval-when-compile
			  (concat "\\b"
				  (regexp-opt '(
"bool" "box" "circle" "char" "char2" "char4" "char8" "char16" "date"
"float4" "float8" "int2" "int4" "int8" "line" "lseg" "money" "path"
"point" "polygon" "serial" "text" "time" "timespan" "timestamp" "varchar"
) t)"\\b")))
	(postgres-builtin-functions (eval-when-compile
			(concat "\\b"
				(regexp-opt '(
;; Misc Postgres builtin functions
"abstime" "age" "area" "box" "center" "date_part" "date_trunc"
"datetime" "dexp" "diameter" "dpow" "float" "float4" "height"
"initcap" "integer" "isclosed" "isfinite" "isoldpath" "isopen"
"length" "lower" "lpad" "ltrim" "pclose" "point" "points" "popen"
"position" "radius" "reltime" "revertpoly" "rpad" "rtrim" "substr"
"substring" "text" "timespan" "translate" "trim" "upgradepath"
"upgradepoly" "upper" "varchar" "width"
) t) "\\b"))))
    (setq mm-mode-postgres-font-lock-keywords
	  (append mm-mode-ansi-font-lock-keywords
		  (list (cons postgres-reserved-words 'font-lock-keyword-face)
			;; XEmacs doesn't have 'font-lock-builtin-face
			(if (string-match "XEmacs\\|Lucid" emacs-version)
			    (cons postgres-builtin-functions 'font-lock-preprocessor-face)
			  ;; Emacs
			  (cons postgres-builtin-functions 'font-lock-builtin-face))
			(cons postgres-types 'font-lock-type-face))))))


(defvar mm-mode-font-lock-keywords mm-mode-ansi-font-lock-keywords
  "MM keywords used by font-lock.

This variable defaults to `mm-mode-ansi-font-lock-keywords'.  This is
used for the default `font-lock-defaults' value in `mm-mode'.  This
can be changed by some entry functions to provide more hilighting.")



;;; Functions to switch highlighting

(defun mm-highlight-oracle-keywords ()
  "Highlight Oracle keywords.
Basically, this just sets `font-lock-keywords' appropriately."
  (interactive)
  (setq font-lock-keywords mm-mode-oracle-font-lock-keywords)
  (font-lock-fontify-buffer))

(defun mm-highlight-postgres-keywords ()
  "Highlight Postgres keywords.
Basically, this just sets `font-lock-keywords' appropriately."
  (interactive)
  (setq font-lock-keywords mm-mode-postgres-font-lock-keywords)
  (font-lock-fontify-buffer))

(defun mm-highlight-ansi-keywords ()
  "Highlight ANSI MM keywords.
Basically, this just sets `font-lock-keywords' appropriately."
  (interactive)
  (setq font-lock-keywords mm-mode-ansi-font-lock-keywords)
  (font-lock-fontify-buffer))



;;; Compatibility functions

(if (not (fboundp 'comint-line-beginning-position))
    ;; comint-line-beginning-position is defined in Emacs 21
    (defun comint-line-beginning-position ()
      "Returns the buffer position of the beginning of the line, after any prompt.
The prompt is assumed to be any text at the beginning of the line matching
the regular expression `comint-prompt-regexp', a buffer local variable."
      (save-excursion (comint-bol nil) (point))))



;;; Small functions

(defun mm-magic-go (arg)
  "Insert \"o\" and call `comint-send-input'.
`mm-electric-stuff' must be the symbol `go'."
  (interactive "P")
  (self-insert-command (prefix-numeric-value arg))
  (if (and (equal mm-electric-stuff 'go)
	   (save-excursion
	     (comint-bol nil)
	     (looking-at "go\\b")))
      (comint-send-input)))

(defun mm-magic-semicolon (arg)
  "Insert semicolon and call `comint-send-input'.
`mm-electric-stuff' must be the symbol `semicolon'."
  (interactive "P")
  (self-insert-command (prefix-numeric-value arg))
  (if (equal mm-electric-stuff 'semicolon)
       (comint-send-input)))

(defun mm-accumulate-and-indent ()
  "Continue MM statement on the next line."
  (interactive)
  (if (fboundp 'comint-accumulate) 
      (comint-accumulate)
    (newline))
  (indent-according-to-mode))

;;;###autoload
(defun mm-help ()
  "Show short help for the MM modes.

Use an entry function to open an interactive MM buffer.  This buffer is
usually named `*MM*'.  The name of the major mode is MMi.

Use the following commands to start a specific MM interpreter:

    PostGres: \\[mm-postgres]
    MyMM: \\[mm-mymm]

Other non-free MM implementations are also supported:

    Solid: \\[mm-solid]
    Oracle: \\[mm-oracle]
    Informix: \\[mm-informix]
    Sybase: \\[mm-sybase]
    Ingres: \\[mm-ingres]
    Microsoft: \\[mm-ms]
    Interbase: \\[mm-interbase]

But we urge you to choose a free implementation instead of these.

Once you have the MMi buffer, you can enter MM statements in the
buffer.  The output generated is appended to the buffer and a new prompt
is generated.  See the In/Out menu in the MMi buffer for some functions
that help you navigate through the buffer, the input history, etc.

If you have a really complex MM statement or if you are writing a
procedure, you can do this in a separate buffer.  Put the new buffer in
`mm-mode' by calling \\[mm-mode].  The name of this buffer can be
anything.  The name of the major mode is MM.

In this MM buffer (MM mode), you can send the region or the entire
buffer to the interactive MM buffer (MMi mode).  The results are
appended to the MMi buffer without disturbing your MM buffer."
  (interactive)
  (describe-function 'mm-help))

(defun mm-read-passwd (prompt &optional default)
  "Read a password using PROMPT.
Optional DEFAULT is password to start with.  This function calls
`read-passwd' if it is available.  If not, function
`ange-ftp-read-passwd' is called.  This should always be available,
even in old versions of Emacs."
  (if (fboundp 'read-passwd)
      (read-passwd prompt nil default)
    (unless (fboundp 'ange-ftp-read-passwd)
      (autoload 'ange-ftp-read-passwd "ange-ftp"))
    (ange-ftp-read-passwd prompt default)))

(defun mm-get-login (&rest what)
  "Get username, password and database from the user.

The variables `mm-user', `mm-password', `mm-server', and
`mm-database' can be customized.  They are used as the default values.
Usernames, servers and databases are stored in `mm-user-history',
`mm-server-history' and `database-history'.  Passwords are not stored
in a history.

Parameter WHAT is a list of the arguments passed to this function.
The function asks for the username if WHAT contains symbol `user', for
the password if it contains symbol `password', for the server if it
contains symbol `server', and for the database if it contains symbol
`database'.

In order to ask the user for username, password and database, call the
function like this: (mm-get-login 'user 'password 'database)."
  (interactive)
  (if (memq 'user what)
      (setq mm-user
	    (read-from-minibuffer "User: " mm-user nil nil
				  mm-user-history)))
  (if (memq 'password what)
      (setq mm-password
	    (mm-read-passwd "Password: " mm-password)))
  (if (memq 'server what)
      (setq mm-server
	    (read-from-minibuffer "Server: " mm-server nil nil
				  mm-server-history)))
  (if (memq 'database what)
      (setq mm-database
	    (read-from-minibuffer "Database: " mm-database nil nil
				  mm-database-history))))

(defun mm-find-mmi-buffer ()
  "Return the current default MMi buffer or nil.
In order to qualify, the MMi buffer must be alive,
be in `mm-interactive-mode' and have a process."
  (let ((default-buffer (default-value 'mm-buffer)))
    (if (and (buffer-live-p default-buffer)
	     (get-buffer-process default-buffer))
	default-buffer
      (save-excursion
	(let ((buflist (buffer-list))
	      (found))
	  (while (not (or (null buflist)
			  found))
	    (let ((candidate (car buflist)))
	      (set-buffer candidate)
	      (if (and (equal major-mode 'mm-interactive-mode)
		       (get-buffer-process candidate))
		  (setq found candidate))
	      (setq buflist (cdr buflist))))
	  found)))))

(defun mm-set-mmi-buffer-generally ()
  "Set MMi buffer for all MM buffers that have none.  
This function checks all MM buffers for their MMi buffer.  If their
MMi buffer is nonexistent or has no process, it is set to the current
default MMi buffer.  The current default MMi buffer is determined
using `mm-find-mmi-buffer'.  If `mm-buffer' is set,
`mm-set-mmi-hook' is run."
  (interactive)
  (save-excursion
    (let ((buflist (buffer-list))
	  (default-mmi-buffer (mm-find-mmi-buffer)))
      (setq-default mm-buffer default-mmi-buffer)
      (while (not (null buflist))
	(let ((candidate (car buflist)))
	  (set-buffer candidate)
	  (if (and (equal major-mode 'mm-mode)
		   (not (buffer-live-p mm-buffer)))
	      (progn
		(setq mm-buffer default-mmi-buffer)
		(run-hooks 'mm-set-mmi-hook))))
	(setq buflist (cdr buflist))))))

(defun mm-set-mmi-buffer ()
  "Set the MMi buffer MM strings are sent to.

Call this function in a MM buffer in order to set the MMi buffer MM
strings are sent to.  Calling this function sets `mm-buffer' and runs
`mm-set-mmi-hook'.

If you call it from a MM buffer, this sets the local copy of
`mm-buffer'.

If you call it from anywhere else, it sets the global copy of
`mm-buffer'."
  (interactive)
  (let ((default-buffer (mm-find-mmi-buffer)))
    (if (null default-buffer)
	(error "There is no suitable MMi buffer"))
    (let ((new-buffer
	   (get-buffer
	    (read-buffer "New MMi buffer: " default-buffer t))))
      (if (null (get-buffer-process new-buffer))
	  (error "Buffer %s has no process" (buffer-name new-buffer)))
      (if (null (save-excursion
		  (set-buffer new-buffer)
		  (equal major-mode 'mm-interactive-mode)))
	  (error "Buffer %s is no MMi buffer" (buffer-name new-buffer)))
      (if new-buffer
	  (progn
	    (setq mm-buffer new-buffer)
	    (run-hooks 'mm-set-mmi-hook))))))

(defun mm-show-mmi-buffer ()
  "Show the name of current MMi buffer.

This is the buffer MM strings are sent to.  It is stored in the
variable `mm-buffer'.  See `mm-help' on how to create such a buffer."
  (interactive)
  (if (null (buffer-live-p mm-buffer))
      (message "%s has no MMi buffer set." (buffer-name (current-buffer)))
    (if (null (get-buffer-process mm-buffer))
	(message "Buffer %s has no process." (buffer-name mm-buffer))
      (message "Current MMi buffer is %s." (buffer-name mm-buffer)))))

(defun mm-make-alternate-buffer-name ()
  "Return a string that can be used to rename a MMi buffer.

This is used to set `mm-alternate-buffer-name' within
`mm-interactive-mode'."
  (concat (if (string= "" mm-user)
	      (if (string= "" (user-login-name))
		  ()
		(concat (user-login-name) "/"))
	    (concat mm-user "/"))
	  (if (string= "" mm-database)
	      (if (string= "" mm-server)
		  (system-name)
		mm-server)
	    mm-database)))

(defun mm-rename-buffer ()
  "Renames a MMi buffer."
  (interactive)
  (rename-buffer (format "*MM: %s*" mm-alternate-buffer-name) t))

(defun mm-copy-column ()
  "Copy current column to the end of buffer.
Inserts SELECT or commas if appropriate."
  (interactive)
  (let ((column))
    (save-excursion
      (setq column (buffer-substring
		  (progn (forward-char 1) (backward-sexp 1) (point))
		  (progn (forward-sexp 1) (point))))
      (goto-char (point-max))
      (let ((bol (comint-line-beginning-position)))
	(cond
	 ;; if empty command line, insert SELECT
	 ((= bol (point))
	  (insert "SELECT "))
	 ;; else if appending to INTO .* (, SELECT or ORDER BY, insert a comma
	 ((save-excursion
	    (re-search-backward "\\b\\(\\(into\\s-+\\S-+\\s-+(\\)\\|select\\|order by\\) .+"
				bol t))
	  (insert ", "))
	 ;; else insert a space
	 (t
	  (if (eq (preceding-char) ? )
	      nil
	    (insert " ")))))
      ;; in any case, insert the column
      (insert column)
      (message "%s" column))))

;; On NT, MM*Plus for Oracle turns on full buffering for stdout if it
;; is not attached to a character device; therefore placeholder
;; replacement by MM*Plus is fully buffered.  The workaround lets
;; Emacs query for the placeholders.

(defvar mm-placeholder-history nil
  "History of placeholder values used.")

(defun mm-query-placeholders-and-send (proc string)
  "Send to PROC input STRING, maybe replacing placeholders.
Placeholders are words starting with and ampersand like &this.
This function is used for `comint-input-sender' if using `mm-oracle' on NT."
  (while (string-match "&\\(\\sw+\\)" string)
    (setq string (replace-match 
		  (read-from-minibuffer
		   (format "Enter value for %s: " (match-string 1 string))
		   nil nil nil mm-placeholder-history)
		  t t string)))
  (comint-send-string proc string)
  (comint-send-string proc "\n"))

;; Using DB2 interactively, newlines must be escaped with " \".
;; The space before the backslash is relevant.
(defun mm-escape-newlines-and-send (proc string)
  "Send to PROC input STRING, escaping newlines if necessary.
Every newline in STRING will be preceded with a space and a backslash."
  (let ((result "") (start 0) mb me)
    (while (string-match "\n" string start)
      (setq mb (match-beginning 0)
	    me (match-end 0))
      (if (and (> mb 1)
	       (string-equal " \\" (substring string (- mb 2) mb)))
	  (setq result (concat result (substring string start me)))
	(setq result (concat result (substring string start mb) " \\\n")))
      (setq start me))
    (setq result (concat result (substring string start)))
    (comint-send-string proc result)
    (comint-send-string proc "\n")))



;;; Sending the region to the MMi buffer.

(defun mm-send-region (start end)
  "Send a region to the MM process."
  (interactive "r")
  (if (buffer-live-p mm-buffer)
      (save-excursion
	(comint-send-region mm-buffer start end)
	(if (string-match "\n$" (buffer-substring start end))
	    ()
	  (comint-send-string mm-buffer "\n"))
	(message "Sent string to buffer %s." (buffer-name mm-buffer))
	(if mm-pop-to-buffer-after-send-region
	    (pop-to-buffer mm-buffer)
	  (display-buffer mm-buffer)))
    (message "No MM process started.")))

(defun mm-send-paragraph ()
  "Send the current paragraph to the MM process."
  (interactive)
  (let ((start (save-excursion
		 (backward-paragraph)
		 (point)))
	(end (save-excursion
	       (forward-paragraph)
	       (point))))
    (mm-send-region start end)))

(defun mm-send-buffer ()
  "Send the buffer contents to the MM process."
  (interactive)
  (mm-send-region (point-min) (point-max)))

(defun mm-toggle-pop-to-buffer-after-send-region (&optional value)
  "Toggle `mm-pop-to-buffer-after-send-region'.

If given the optional parameter VALUE, sets
mm-toggle-pop-to-buffer-after-send-region to VALUE."
  (interactive "P")
  (if value
      (setq mm-pop-to-buffer-after-send-region value)
    (setq mm-pop-to-buffer-after-send-region
	  (null mm-pop-to-buffer-after-send-region ))))



;;; MM mode -- uses MM interactive mode


;;(defconst c-symbol-key "[_a-zA-Z]\\(\\w\\|\\s_\\)*")
;;(defconst c-mm-operator-key "operator")
;;(defvar c-class-key c-mm-operator-key)
;;(make-variable-buffer-local 'c-class-key)

;;;###autoload
(defun mm-mode ()
  "Major mode to edit MM.

You can send MM statements to the MMi buffer using
\\[mm-send-region].  Such a buffer must exist before you can do this.
See `mm-help' on how to create MMi buffers.

\\{mm-mode-map}
Customization: Entry to this mode runs the `mm-mode-hook'.

When you put a buffer in MM mode, the buffer stores the last MMi
buffer created as its destination in the variable `mm-buffer'.  This
will be the buffer \\[mm-send-region] sends the region to.  If this
MMi buffer is killed, \\[mm-send-region] is no longer able to
determine where the strings should be sent to.  You can set the
value of `mm-buffer' using \\[mm-set-mmi-buffer].

For information on how to create multiple MMi buffers, see
`mm-interactive-mode'.

Note that MM doesn't really have an escape character.  If you want
to use the backslash as an escape character, you must tell Emacs.
Here's how to do that in your ~/.emacs file:

\(add-hook 'mm-mode-hook
          \(lambda ()
	    \(modify-syntax-entry ?\\\\ \".\" mm-mode-syntax-table)))"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'mm-mode)
  (setq mode-name "MM")
  (use-local-map mm-mode-map)
  (if mm-mode-menu
      (easy-menu-add mm-mode-menu)); XEmacs
  (set-syntax-table mm-mode-syntax-table)
  (make-local-variable 'font-lock-defaults)
  ;; Note that making KEYWORDS-ONLY nil will cause havoc if you try
  ;; SELECT 'x' FROM DUAL with MM*Plus, because the title of the column
  ;; will have just one quote.  Therefore syntactic hilighting is
  ;; disabled for interactive buffers.  `_' and `.' are considered part
  ;; of words.
  (setq font-lock-defaults '(mm-mode-font-lock-keywords
			     nil t ((?_ . "w") (?. . "w"))))
  (make-local-variable 'comment-start)
  (setq comment-start "/[/*]")
  (make-local-variable 'c-class-key)
  (setq c-class-key "operator")
  
  ;; Make each buffer in mm-mode remember the "current" MMi buffer.
  (make-local-variable 'mm-buffer)
  ;; Add imenu support for mm-mode.  Note that imenu-generic-expression
  ;; is buffer-local, so we don't need a local-variable for it.  MM is
  ;; case-insensitive, that's why we have to set imenu-case-fold-search.
  ;; imenu-syntax-alist makes sure that `_' is considered part of object
  ;; names.
  (setq imenu-generic-expression mm-imenu-generic-expression
	imenu-case-fold-search t
	imenu-syntax-alist '(("_" . "w")))
  ;; Make `mm-send-paragraph' work on paragraphs that contain indented
  ;; lines.
  (make-local-variable 'paragraph-separate)
  (make-local-variable 'paragraph-start)
  (setq paragraph-separate "[\f]*$"
	paragraph-start "[\n\f]")
  ;; Abbrevs
  (setq local-abbrev-table mm-mode-abbrev-table)
  (setq abbrev-all-caps 1)
  ;; Run hook
  (run-hooks 'mm-mode-hook))



;;; MM interactive mode

(put 'mm-interactive-mode 'mode-class 'special)

(defun mm-interactive-mode ()
  "Major mode to use a MM interpreter interactively.

Do not call this function by yourself.  The environment must be
initialized by an entry function specific for the MM interpreter.  See
`mm-help' for a list of available entry functions.

\\[comint-send-input] after the end of the process' output sends the
text from the end of process to the end of the current line.
\\[comint-send-input] before end of process output copies the current
line minus the prompt to the end of the buffer and sends it.
\\[comint-copy-old-input] just copies the current line.
Use \\[mm-accumulate-and-indent] to enter multi-line statements.

If you want to make multiple MM buffers, rename the `*MM*' buffer
using \\[rename-buffer] or \\[rename-uniquely] and start a new process.
See `mm-help' for a list of available entry functions.  The last buffer
created by such an entry function is the current MMi buffer.  MM
buffers will send strings to the MMi buffer current at the time of
their creation.  See `mm-mode' for details.

Sample session using two connections:

1. Create first MMi buffer by calling an entry function.
2. Rename buffer \"*MM*\" to \"*Connection 1*\".
3. Create a MM buffer \"test1.mm\".
4. Create second MMi buffer by calling an entry function.
5. Rename buffer \"*MM*\" to \"*Connection 2*\".
6. Create a MM buffer \"test2.mm\".

Now \\[mm-send-region] in buffer \"test1.mm\" will send the region to
buffer \"*Connection 1*\", \\[mm-send-region] in buffer \"test2.mm\"
will send the region to buffer \"*Connection 2*\".

If you accidentally suspend your process, use \\[comint-continue-subjob]
to continue it.  On some operating systems, this will not work because
the signals are not supported.

\\{mm-interactive-mode-map}
Customization: Entry to this mode runs the hooks on `comint-mode-hook'
and `mm-interactive-mode-hook' (in that order).  Before each input, the
hooks on `comint-input-filter-functions' are run.  After each MM
interpreter output, the hooks on `comint-output-filter-functions' are
run.

Variable `mm-input-ring-file-name' controls the initialisation of the
input ring history.

Variables `comint-output-filter-functions', a hook, and
`comint-scroll-to-bottom-on-input' and
`comint-scroll-to-bottom-on-output' control whether input and output
cause the window to scroll to the end of the buffer.

If you want to make MM buffers limited in length, add the function
`comint-truncate-buffer' to `comint-output-filter-functions'.

Here is an example for your .emacs file.  It keeps the MMi buffer a
certain length.

\(add-hook 'mm-interactive-mode-hook
    \(function (lambda ()
        \(setq comint-output-filter-functions 'comint-truncate-buffer))))

Here is another example.  It will always put point back to the statement
you entered, right above the output it created.

\(setq comint-output-filter-functions
       \(function (lambda (STR) (comint-show-output))))"
  (comint-mode)
  (setq comint-prompt-regexp mm-prompt-regexp)
  (setq left-margin mm-prompt-length)
  (setq major-mode 'mm-interactive-mode)
  (setq mode-name "MMi")
  (use-local-map mm-interactive-mode-map)
  (if mm-interactive-mode-menu
      (easy-menu-add mm-interactive-mode-menu)); XEmacs
  (set-syntax-table mm-mode-syntax-table)
  (make-local-variable 'font-lock-defaults)
  ;; Note that making KEYWORDS-ONLY nil will cause havoc if you try
  ;; SELECT 'x' FROM DUAL with MM*Plus, because the title of the column
  ;; will have just one quote.  Therefore syntactic hilighting is
  ;; disabled for interactive buffers.  `_' and `.' are considered part
  ;; of words.
  (setq font-lock-defaults '(mm-mode-font-lock-keywords
			     t t ((?_ . "w") (?. . "w"))))
  ;; Enable commenting and uncommenting of the region.
  (make-local-variable 'comment-start)
  (setq comment-start "/[/*]")
  (make-local-variable 'c-class-key)
  (setq c-class-key "operator")

  ;; Abbreviation table init and case-insensitive.  It is not activatet
  ;; by default.
  (setq local-abbrev-table mm-mode-abbrev-table)
  (setq abbrev-all-caps 1)
  ;; Exiting the process will call mm-stop.
  (set-process-sentinel (get-buffer-process mm-buffer) 'mm-stop)
  ;; People wanting a different history file for each
  ;; buffer/process/client/whatever can change separator and file-name
  ;; on the mm-interactive-mode-hook.
  (setq comint-input-ring-separator mm-input-ring-separator
	comint-input-ring-file-name mm-input-ring-file-name)
  ;; Create a usefull name for renaming this buffer later.
  (make-local-variable 'mm-alternate-buffer-name)
  (setq mm-alternate-buffer-name (mm-make-alternate-buffer-name))
  ;; User stuff.
  (run-hooks 'mm-interactive-mode-hook)
  ;; Calling the hook before calling comint-read-input-ring allows users
  ;; to set comint-input-ring-file-name in mm-interactive-mode-hook.
  (comint-read-input-ring t))

(defun mm-stop (process event)
  "Called when the MM process is stopped.

Writes the input history to a history file using
`comint-write-input-ring' and inserts a short message in the MM buffer.

This function is a sentinel watching the MM interpreter process.
Sentinels will always get the two parameters PROCESS and EVENT."
  (comint-write-input-ring)
  (if (and (eq (current-buffer) mm-buffer)
	   (not buffer-read-only))
      (insert (format "\nProcess %s %s\n" process event))
    (message "Process %s %s" process event)))



;;; Entry functions for different MM interpreters.

;;;###autoload
(defun mm-oracle ()
  "Run mmplus by Oracle as an inferior process.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-oracle-program'.  Login uses
the variables `mm-user', `mm-password', and `mm-database' as
defaults, if set.  Additional command line parameters can be stored in
the list `mm-oracle-options'.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-oracle].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (mm-get-login 'user 'password 'database)
    (message "Login...")
    ;; Produce user/password@database construct.  Password without user
    ;; is meaningless; database without user/password is meaningless,
    ;; because "@param" will ask mmplus to interpret the script
    ;; "param".
    (let ((parameter nil))
      (if (not (string= "" mm-user))
	  (if (not (string= "" mm-password))
	      (setq parameter (concat mm-user "/" mm-password))
	    (setq parameter mm-user)))
      (if (and parameter (not (string= "" mm-database)))
	  (setq parameter (concat parameter "@" mm-database)))
      (if parameter
	  (setq parameter (nconc (list parameter) mm-oracle-options))
	(setq parameter mm-oracle-options))
      (if parameter
	  (set-buffer (apply 'make-comint "MM" mm-oracle-program nil 
			     parameter))
	(set-buffer (make-comint "MM" mm-oracle-program nil))))
    (setq mm-prompt-regexp "^MM> ")
    (setq mm-prompt-length 5)
    (setq mm-buffer (current-buffer))
    ;; set mm-mode-font-lock-keywords to something different before
    ;; calling mm-interactive-mode.
    (setq mm-mode-font-lock-keywords mm-mode-oracle-font-lock-keywords)
    (mm-interactive-mode)
    ;; If running on NT, make sure we do placeholder replacement
    ;; ourselves.  This must come after mm-interactive-mode because all
    ;; local variables will be killed, there.
    (if (eq window-system 'w32)
	(setq comint-input-sender 'mm-query-placeholders-and-send))
    (message "Login...done")
    (pop-to-buffer mm-buffer)))



;;;###autoload
(defun mm-sybase ()
  "Run imm by SyBase as an inferior process.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-sybase-program'.  Login uses
the variables `mm-server', `mm-user', `mm-password', and
`mm-database' as defaults, if set.  Additional command line parameters
can be stored in the list `mm-sybase-options'.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-sybase].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (mm-get-login 'server 'user 'password 'database)
    (message "Login...")
    ;; Put all parameters to the program (if defined) in a list and call
    ;; make-comint.
    (let ((params mm-sybase-options))
      (if (not (string= "" mm-server))
	  (setq params (append (list "-S" mm-server) params)))
      (if (not (string= "" mm-database))
	  (setq params (append (list "-D" mm-database) params)))
      (if (not (string= "" mm-password))
	  (setq params (append (list "-P" mm-password) params)))
      (if (not (string= "" mm-user))
	  (setq params (append (list "-U" mm-user) params)))
      (set-buffer (apply 'make-comint "MM" mm-sybase-program
			 nil params)))
    (setq mm-prompt-regexp "^MM> ")
    (setq mm-prompt-length 5)
    (setq mm-buffer (current-buffer))
    (mm-interactive-mode)
    (message "Login...done")
    (pop-to-buffer mm-buffer)))



;;;###autoload
(defun mm-informix ()
  "Run dbaccess by Informix as an inferior process.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-informix-program'.  Login uses
the variable `mm-database' as default, if set.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-informix].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (mm-get-login 'database)
    (message "Login...")
    ;; username and password are ignored.
    (if (string= "" mm-database)
	(set-buffer (make-comint "MM" mm-informix-program nil))
      (set-buffer (make-comint "MM" mm-informix-program nil mm-database "-")))
    (setq mm-prompt-regexp "^MM> ")
    (setq mm-prompt-length 5)
    (setq mm-buffer (current-buffer))
    (mm-interactive-mode)
    (message "Login...done")
    (pop-to-buffer mm-buffer)))



;;;###autoload
(defun mm-mymm ()
  "Run mymm by TcX as an inferior process.

Mymm versions 3.23 and up are free software.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-mymm-program'.  Login uses
the variables `mm-user', `mm-password', `mm-database', and
`mm-server' as defaults, if set.  Additional command line parameters
can be stored in the list `mm-mymm-options'.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-mymm].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (mm-get-login 'user 'password 'database 'server)
    (message "Login...")
    ;; Put all parameters to the program (if defined) in a list and call
    ;; make-comint.
    (let ((params))
      (if (not (string= "" mm-database))
	  (setq params (append (list mm-database) params)))
      (if (not (string= "" mm-server))
	  (setq params (append (list (concat "--host=" mm-server)) params)))
      (if (not (string= "" mm-password))
	  (setq params (append (list (concat "--password=" mm-password)) params)))
      (if (not (string= "" mm-user))
	  (setq params (append (list (concat "--user=" mm-user)) params)))
      (if (not (null mm-mymm-options))
          (setq params (append mm-mymm-options params)))
      (set-buffer (apply 'make-comint "MM" mm-mymm-program
			 nil params)))
    (setq mm-prompt-regexp "^mymm>")
    (setq mm-prompt-length 6)
    (setq mm-buffer (current-buffer))
    (mm-interactive-mode)
    (message "Login...done")
    (pop-to-buffer mm-buffer)))



;;;###autoload
(defun mm-solid ()
  "Run solmm by Solid as an inferior process.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-solid-program'.  Login uses
the variables `mm-user', `mm-password', and `mm-server' as
defaults, if set.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-solid].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (mm-get-login 'user 'password 'server)
    (message "Login...")
    ;; Put all parameters to the program (if defined) in a list and call
    ;; make-comint.
    (let ((params))
      ;; It only makes sense if both username and password are there.
      (if (not (or (string= "" mm-user)
		   (string= "" mm-password)))
	  (setq params (append (list mm-user mm-password) params)))
      (if (not (string= "" mm-server))
	  (setq params (append (list mm-server) params)))
      (set-buffer (apply 'make-comint "MM" mm-solid-program
			 nil params)))
    (setq mm-prompt-regexp "^")
    (setq mm-prompt-length 0)
    (setq mm-buffer (current-buffer))
    (mm-interactive-mode)
    (message "Login...done")
    (pop-to-buffer mm-buffer)))



;;;###autoload
(defun mm-ingres ()
  "Run mm by Ingres as an inferior process.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-ingres-program'.  Login uses
the variable `mm-database' as default, if set.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-ingres].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (mm-get-login 'database)
    (message "Login...")
    ;; username and password are ignored.
    (if (string= "" mm-database)
	(set-buffer (make-comint "MM" mm-ingres-program nil))
      (set-buffer (make-comint "MM" mm-ingres-program nil mm-database)))
    (setq mm-prompt-regexp "^\* ")
    (setq mm-prompt-length 2)
    (setq mm-buffer (current-buffer))
    (mm-interactive-mode)
    (message "Login...done")
    (pop-to-buffer mm-buffer)))



;;;###autoload
(defun mm-ms ()
  "Run imm by Microsoft as an inferior process.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-ms-program'.  Login uses the
variables `mm-user', `mm-password', `mm-database', and `mm-server'
as defaults, if set.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-ms].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (mm-get-login 'user 'password 'database 'server)
    (message "Login...")
    ;; Put all parameters to the program (if defined) in a list and call
    ;; make-comint.
    (let ((params '("-w 300")))
      (if (not (string= "" mm-server))
        (setq params (append (list "-S" mm-server) params)))
      (if (not (string= "" mm-database))
        (setq params (append (list "-d" mm-database) params)))
      (if (not (string= "" mm-user))
	  (setq params (append (list "-U" mm-user) params)))
      (if (not (string= "" mm-password))
	  (setq params (append (list "-P" mm-password) params))
	;; If -P is passed to IMM as the last argument without a password,
	;; it's considered null.
	(setq params (append params (list "-P"))))
      (set-buffer (apply 'make-comint "MM" mm-ms-program
			 nil params)))
    (setq mm-prompt-regexp "^[0-9]*>")
    (setq mm-prompt-length 5)
    (setq mm-buffer (current-buffer))
    (mm-interactive-mode)
    (message "Login...done")
    (pop-to-buffer mm-buffer)))



;;;###autoload
(defun mm-postgres ()
  "Run pmm by Postgres as an inferior process.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-postgres-program'.  Login uses
the variables `mm-database' and `mm-server' as default, if set.
Additional command line parameters can be stored in the list
`mm-postgres-options'.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-postgres].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.  If your output lines end with ^M,
your might try undecided-dos as a coding system.  If this doesn't help,
Try to set `comint-output-filter-functions' like this:

\(setq comint-output-filter-functions (append comint-output-filter-functions
					     '(comint-strip-ctrl-m)))

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (mm-get-login 'database 'server)
    (message "Login...")
    ;; username and password are ignored.  Mark Stosberg suggest to add
    ;; the database at the end.  Jason Beegan suggest using --pset and
    ;; pager=off instead of \\o|cat.  The later was the solution by
    ;; Gregor Zych.  Jason's suggestion is the default value for
    ;; mm-postgres-options.
    (let ((params mm-postgres-options))
      (if (not (string= "" mm-database))
	  (setq params (append params (list mm-database))))
      (if (not (string= "" mm-server))
	  (setq params (append (list "-h" mm-server) params)))
      (set-buffer (apply 'make-comint "MM" mm-postgres-program
			 nil params)))
    (setq mm-prompt-regexp "^.*> *")
    (setq mm-prompt-length 5)
    ;; This is a lousy hack to prevent pmm from truncating it's output
    ;; and giving stupid warnings. If s.o. knows a way to prevent pmm
    ;; from acting this way, then I would be very thankful to
    ;; incorporate this (Gregor Zych <zych@pool.informatik.rwth-aachen.de>)
    ;; (comint-send-string "*MM*" "\\o \| cat\n")
    (setq mm-mode-font-lock-keywords mm-mode-postgres-font-lock-keywords)
    (setq mm-buffer (current-buffer))
    (mm-interactive-mode)
    (message "Login...done")
    (pop-to-buffer mm-buffer)))



;;;###autoload
(defun mm-interbase ()
  "Run imm by Interbase as an inferior process.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-interbase-program'.  Login
uses the variables `mm-user', `mm-password', and `mm-database' as
defaults, if set.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-interbase].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (mm-get-login 'user 'password 'database)
    (message "Login...")
    ;; Put all parameters to the program (if defined) in a list and call
    ;; make-comint.
    (let ((params mm-interbase-options))
      (if (not (string= "" mm-user))
	  (setq params (append (list "-u" mm-user) params)))
      (if (not (string= "" mm-password))
	  (setq params (append (list "-p" mm-password) params)))
      (if (not (string= "" mm-database))
        (setq params (cons mm-database params))); add to the front!
      (set-buffer (apply 'make-comint "MM" mm-interbase-program
			 nil params)))
    (setq mm-prompt-regexp "^MM> ")
    (setq mm-prompt-length 5)
    (setq mm-buffer (current-buffer))
    (mm-interactive-mode)
    (message "Login...done")
    (pop-to-buffer mm-buffer)))



;;;###autoload
(defun mm-db2 ()
  "Run db2 by IBM as an inferior process.

If buffer `*MM*' exists but no process is running, make a new process.
If buffer exists and a process is running, just switch to buffer
`*MM*'.

Interpreter used comes from variable `mm-db2-program'.  There is not
automatic login.

The buffer is put in mm-interactive-mode, giving commands for sending
input.  See `mm-interactive-mode'.

If you use \\[mm-accumulate-and-indent] to send multiline commands to
db2, newlines will be escaped if necessary.  If you don't want that, set
`comint-input-sender' back to `comint-simple-send' by writing an after
advice.  See the elisp manual for more information.

To specify a coding system for converting non-ASCII characters
in the input and output to the process, use \\[universal-coding-system-argument]
before \\[mm-db2].  You can also specify this with \\[set-buffer-process-coding-system]
in the MM buffer, after you start the process.
The default comes from `process-coding-system-alist' and
`default-process-coding-system'.

\(Type \\[describe-mode] in the MM buffer for a list of commands.)"
  (interactive)
  (if (comint-check-proc "*MM*")
      (pop-to-buffer "*MM*")
    (message "Login...")
    ;; Put all parameters to the program (if defined) in a list and call
    ;; make-comint.
    (set-buffer (apply 'make-comint "MM" mm-db2-program
		       nil mm-db2-options))
    (setq mm-prompt-regexp "^db2 => ")
    (setq mm-prompt-length 7)
    (setq mm-buffer (current-buffer))
    (mm-interactive-mode)
    ;; Escape newlines.  This must come after mm-interactive-mode
    ;; because all local variables will be killed, there.
    (setq comint-input-sender 'mm-escape-newlines-and-send)
    (message "Login...done")
    (pop-to-buffer mm-buffer)))

(provide 'mm)

;;; mm.el ends here
