/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package edu.stanford.db.rdf.query.memory;

import org.w3c.rdf.model.*;
import edu.stanford.db.rdf.query.*;
import org.w3c.rdf.util.RDFUtil;

import java.util.*;

/**
 * This is a query iterator.
 *
 * @author Sergey Melnik <melnik@db.stanford.edu> */

public class JoinIterator implements QueryIterator {
    
    QueryIterator[] plan;
    int current = -1;
    // hold preset variables (do not reset them)
    Hashtable presetVars = new Hashtable();
    boolean empty = false;
    // first occurrence of the variable
    Hashtable claimed = new Hashtable(); // varName->iterator

    // we only need model to be able to pass it to underlying
    // iterators, it is not used otherwise
    Model model;
    int cost;

  public QueryIterator duplicate() {

    JoinIterator it = new JoinIterator();
    QueryIterator plan[] = new QueryIterator[this.plan.length];
    for(int i = 0; i < plan.length; i++) {
      plan[i] = (QueryIterator)this.plan[i].duplicate();
    }
    it.plan = plan;
    it.claimed = claimed; // don't clone, this is a constant
    return it;
  }

  JoinIterator() {
  }

  public void optimize() throws ModelException {

    // TODO: here we could optimize (reorder) the plan if necessary
    // make sure that at least one variable in the next iterator has been bound in previous iterator

    // start with the last iterator, bubble it up as far as necessary
    // X,Y
    // Z,W
    // Y,Z

    Set[] vars = getVarSets();
    //    println("Current varsets: " + Arrays.asList(vars));

    // start with a leaf iterator, i.e. the one that contains at least one leaf variable
    Set leafs = new HashSet();
    Set nonLeafs = new HashSet();
    // count variables
    int[] varcount = new int[claimed.size()];
    for(int i=0; i < vars.length; i++) {
      for(Iterator it = vars[i].iterator(); it.hasNext();) {
	Object var = it.next();
	if(!leafs.add(var)) { // non-unique!
	  nonLeafs.add(var);
	}
      }
    }
    leafs.removeAll(nonLeafs);

    // move up iterator that contains a leaf var
    int leafIdx = 0;
    while(leafIdx < plan.length)
      if(overlap(vars[leafIdx], leafs))
	break;
      else
	leafIdx++;
    if(leafIdx > 0 && leafIdx < plan.length)
      swap(vars, plan, 0, leafIdx);

    // make sure each two consecutive iterators have overlapping variables
    // start with second
    Set ivars = new HashSet();

    for(int i=1; i < plan.length; i++) {

      ivars.addAll(vars[i-1]);
      int j = i;
      while(j < plan.length && !overlap(vars[j], ivars))
	j++;
      if(j > i && j < plan.length)
	swap(vars, plan, i, j);
    }
    //    println("Optimized varsets: " + Arrays.asList(vars));

    // rebuild the map
    generateVariableMap();
  }

  static void swap(Set[] vars, QueryIterator[] plan, int i, int j) {

    // exchange i with some other item further down
    Object tmp = vars[i];
    vars[i] = vars[j];
    vars[j] = (Set)tmp;
    tmp = plan[i];
    plan[i] = plan[j];
    plan[j] = (QueryIterator)tmp;
  }

  Set[] getVarSets() throws ModelException {

    Set[] vars = new Set[plan.length]; // holds a list of ordered variable lists
    for(int i=0; i < plan.length; i++) {

      Set l = new HashSet();
      Enumeration en = plan[i].getVariables();
      while(en.hasMoreElements())
	l.add(en.nextElement());
//       Object[] a = l.toArray();
//       Arrays.sort(a);
      vars[i] = l;
    }
    return vars;
  }

    static boolean overlap(Set a, Set b) {

      for(Iterator it = a.iterator(); it.hasNext();)
	if(b.contains(it.next()))
	  return true;
      return false;
    }

    public JoinIterator(QueryIterator[] plan)  {
	
	this.plan = plan;
	//	optimize();

	generateVariableMap();
    }

  void generateVariableMap() {

    claimed.clear();
    // determine first occurrences of variables
    try {
      for(int i = 0; i < plan.length; i++) {

	QueryIterator it = plan[i];

	for(Enumeration en = it.getVariables(); en.hasMoreElements();) {

	  String varName = (String)en.nextElement();
	  if(varName != null && !claimed.containsKey(varName)) {
	    claimed.put(varName, new Integer(i));
	    //	    println("Claimed " + varName + " for " + i);
	  }
	}
      }
    } catch (ModelException any) {} // can only happen in getVariables, but unlikely
  }

    private void println(String msg) {

      System.err.println(toString() + ": " + msg);
    }

    public Enumeration getVariables() {

	return claimed.keys();
    }

    public int getCost() {

	int total = cost;
	for(int i = 0; i < plan.length; i++)
	    total += plan[i].getCost();
	return total;
    }


    // can be speeded up by precomputing inverse table
    void resetVariables(int it, Hashtable vars) {

      //      println("Resetting " + claimed.size() + " variables");
	for(Enumeration en = claimed.keys(); en.hasMoreElements();) {

	    String var = (String)en.nextElement();
	    int i = ((Integer)claimed.get(var)).intValue();
	    //	    println("Reset var? " + var + " claimed by " + i + " at " + it);
	    if(i == it) {
		RDFNode n = (RDFNode)presetVars.get(var);
		if(n == null) {
		    vars.remove(var);
		    //		    println("Reset var: " + var);
		}
	    }
	}
    }
    
    public void reset(Hashtable vars, Model m) throws ModelException {

      //        println("Resetting, having " + claimed.size() + " claimed variables");
	current = -1;
	presetVars = (Hashtable)vars.clone();
	empty = false;
	if(m != null)
	    model = m;

// 	for(int i=0; i < plan.length; i++)
// 	  plan[i].reset(vars, m);
    }

    public boolean next(Hashtable vars) throws ModelException {
	
	if(empty)
	    return false;

	//	println("Current: " + current);
	
	// if current iterator to advance is the last one, just advance the last one
	if(current == plan.length-1) {
	    if(plan[current].next(vars)) {
		cost++;
		return true;
	    }
	}

	// rewind forward if current < plan.length-1
	while(true) {

	    boolean found = true;

	    if(current < plan.length-1) {
		while(current+1 < plan.length && found) {
		  //		    println("Reset: " + (current+1));
		    resetVariables(++current, vars);
		    plan[current].reset(vars, model);
		    found &= plan[current].next(vars);
		    cost++;
		}
		if(found)
		    return true;
	    }
	    
	    // else backtrack
	    found = false;
	    while(current > 0 && 
		  (cost++ >= 0) && // count only
		  !(found = plan[--current].next(vars)));
	    //	    println("Back to: " + current);
	    if(!found && current == 0) {
		// no more left
		empty = true;
		return false;
	    } else
		continue;
	}
    }
}
