package edu.stanford.db.rdf.util;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import java.util.*;
import edu.stanford.db.rdf.vocabulary.order_20000527.RDFX;
import edu.stanford.db.rdf.query.*;
import edu.stanford.db.rdf.query.memory.*;
import edu.stanford.db.rdf.query.orderings.*;

/**
 * This class provides a number of convenience methods for model access
 */

public class PowerModel implements Model {

  Model m;
  
  public static PowerModel create(Model m) {

    if(m instanceof PowerModel)
      return (PowerModel)m;
    else
      return new PowerModel(m);
  }

  protected PowerModel(Model m) {

    this.m = m;
  }

  public Model getPeer() {

    return m;
  }

  /**
   * The formal string label of the node.
   * URI in case of a resource, string in case of a literal.
   */
  public String getLabel() throws ModelException {

    return m.getLabel();
  }


  /**
   * Returns the URI of the resource. Triples and models must implement this method in a standard way.
   * @return the URI of the resource
   *
   * @see org.w3c.rdf.model.Statement
   * @see org.w3c.rdf.model.Model
   */
  public String getURI() throws ModelException {

    return m.getURI();
  }

  /**
   * Returns the namespace of the resource. May return null.
   * @return the namespace of the resource
   *
   * @see org.w3c.rdf.model.Statement
   * @see org.w3c.rdf.model.Model
   * @since 2000-10-21
   */
  public String getNamespace() throws ModelException {

    return m.getNamespace();
  }

  /**
   * Returns the local name of the resource. May not return null.
   * @return the local name of the resource
   *
   * @see org.w3c.rdf.model.Statement
   * @see org.w3c.rdf.model.Model
   * @since 2000-10-21
   */
  public String getLocalName() throws ModelException {

    return m.getLocalName();
  }

  // Model interface


  /**
   * Set a base URI for the model.
   */
  public void setSourceURI(String uri) throws ModelException {

    m.setSourceURI(uri);
  }

  /**
   * Returns current base URI setting.
   */
  public String getSourceURI() throws ModelException {

    return m.getSourceURI();
  }

  // Model access

  /**
   * Number of triples in the model
   *
   * @return  number of triples, -1 if unknown
   * 
   * @seeAlso org.w3c.rdf.model.VirtualModel
   */
  public int size() throws ModelException {

    return m.size();
  }

  /**
   * true if the model contains no triples
   */
  public boolean isEmpty() throws ModelException {

    return m.isEmpty();
  }

  /**
   * Enumerate triples
   */
  public Enumeration elements() throws ModelException {

    return m.elements();
  }

  /**
   * Tests if the model contains the given triple.
   *
   * @return  <code>true</code> if the triple belongs to the model;
   *          <code>false</code> otherwise.
   */
  public boolean contains(Statement t) throws ModelException {

    return m.contains(t);
  }

  // Model manipulation: add, remove, find

  /**
   * Adds a new triple to the model.
   */
  public void add(Statement t) throws ModelException {

    m.add(t);
  }

  /**
   * Removes the triple from the model.
   */
  public void remove(Statement t) throws ModelException {

    m.remove(t);
  }

  /**
   * True if the model supports <tt>add()</tt> and <tt>remove()</tt> methods.
   * A model may change behavior of this function over time.
   */
  public boolean isMutable() throws ModelException {

    return m.isMutable();
  }

  /**
   * General method to search for triples.
   * <code>null</code> input for any parameter will match anything.
   * <p>Example: <code>Model result = m.find( null, RDF.type, m.getNodeFactory().createResource("http://...#MyClass") )</code>
   * <p>finds all instances of the class <code>MyClass</code>
   */
  public Model find( Resource subject, Resource predicate, RDFNode object ) throws ModelException {

    return m.find(subject, predicate, object);
  }

  public Statement find1(Resource s, Resource p, RDFNode o) throws ModelException {

    return RDFUtil.get1(m.find(s, p, o));
  }

  /**
   * Search for a single triple.
   * <code>null</code> input for any parameter will match anything.
   *
   * @return  a single triple if such was found;<br>
   *          <code>null</code> if nothing was found;<br>
   *          otherwise a RuntimeException is thrown.
  public Statement find1( Resource subject, Resource predicate, RDFnode object );
   */

  /**
   * Clone the model.
   */
  public Model duplicate() throws ModelException {

    return m.duplicate();
  }

  /**
   * Creates empty model of the same Class
   */
  public Model create() throws ModelException {

    return m.create();
  }

  /**
   * Returns the node factory for this model
   */
  public NodeFactory getNodeFactory() throws ModelException {

    return m.getNodeFactory();
  }

  // CONVENIENCE METHODS

  public Statement add(Resource s, Resource p, String o) throws ModelException {
    return add(s, p, literal(o));
  }

  public Statement add(Resource s, Resource p, RDFNode o) throws ModelException {

    Statement st = m.getNodeFactory().createStatement(s, p, o);
    m.add(st);
    return st;
  }

  public Statement statement(Resource s, Resource p, RDFNode o) throws ModelException {

    return m.getNodeFactory().createStatement(s, p, o);
  }

  public Resource resource(String s) throws ModelException {

    return m.getNodeFactory().createResource(s);
  }

  public Literal literal(String s) throws ModelException {

    return m.getNodeFactory().createLiteral(s);
  }

  public Literal literal(int i) throws ModelException {

    return m.getNodeFactory().createLiteral(i);
  }

  public Resource resource(String ns, String s) throws ModelException {

    return m.getNodeFactory().createResource(ns, s);
  }

  public void addOrdered(Resource s, Resource p, List list) throws ModelException {

    addOrdered(s, p, list.iterator());
    /*
    for(int i=0; i < list.size(); i++) {

      Object obj = list.get(i);
      add(s, p,  obj instanceof String ? (RDFNode)literal((String)obj) : (RDFNode)obj, i+1);
    }
    */
  }

  public void addOrdered(Resource s, Resource p, Iterator it) throws ModelException {

    int i=0;
    while(it.hasNext()) {
      i++;
      Object obj = it.next();
      add(s, p,  obj instanceof String ? (RDFNode)literal((String)obj) : (RDFNode)obj, i);
    }
  }

  public void add(Resource s, Resource p, RDFNode o, int pos) throws ModelException {

    Statement st = add(s, p, o);
    if(pos > 0)
      add(st, RDFX.order, String.valueOf(pos));
  }

  public void addLast(Resource s, Resource p, RDFNode o) throws ModelException {

    // find last position
    List l = getOrderedList(s, p);
    add(s, p, o, l.size() + 1);
  }

  /**
   * iterator variables are: "S", "P", "O", and "ST"; "ORD", "ST2"
   */
  public NodeIterator getOrderedNI2(Resource s, Resource p) throws ModelException {

    QueryIterator[] q1 = new QueryIterator[] { new FindIterator("S", "P", "O", "ST"),
					       new OptionalIterator(new FindIterator("ST", RDFX.order, "ORD", "ST2")) };

    QueryIterator q = new OrderIterator(new JoinIterator(q1),
					new Numeric("ORD"));

    NodeIterator it = new NodeResultSet(q, m);
    it.set("S", s);
    it.set("P", p);

    return it;
  }

  /**
   * iterator variables are: "S", "P", "O", and "ST"
   */
  public NodeIterator getOrderedNI(Resource s, Resource p) throws ModelException {

    QueryIterator q = new OrderIterator(new FindIterator("S", "P", "O", "ST"),
					new ByOrderProperty("ST"));

    NodeIterator it = new NodeResultSet(q, m);
    it.set("S", s);
    it.set("P", p);

    return it;
  }

  public Iterator getOrderedIterator(Resource s, Resource p) throws ModelException {

    return getOrderedList(s, p).iterator();
  }
  
  public List getOrderedList(Resource s, Resource p) throws ModelException {

    List l = new ArrayList();
    for(NodeIterator it = getOrderedNI(s, p); it.next();) {

      l.add(it.get("O"));
    }
    return l;
  }

  public Resource subjectR(Resource predicate, RDFNode object) throws ModelException {

    return RDFUtil.getSubject(this, predicate, object);
  }
  
  public Resource objectR(Resource subject, Resource predicate) throws ModelException {

    return RDFUtil.getObjectResource(this, subject, predicate);
  }
  
  public String objectS(Resource subject, Resource predicate) throws ModelException {

    return RDFUtil.getObjectLiteral(this, subject, predicate);
  }

  public Literal objectL(Resource subject, Resource predicate) throws ModelException {

    RDFNode n = object(subject, predicate);
    if(n instanceof Literal)
      return (Literal)n;
    return null;
  }

  public int objectInt(Resource subject, Resource predicate) throws ModelException {

    return asInt(object(subject, predicate));
  }

  public int asInt(RDFNode n) throws ModelException {

    if(n instanceof Literal)
      try {
	return Integer.parseInt(n.getLabel());
      } catch(NumberFormatException nfe) {}

    return Integer.MIN_VALUE;
  }

  public RDFNode object(Resource subject, Resource predicate) throws ModelException {

    return RDFUtil.getObject(this, subject, predicate);
  }

  public int count(Resource s, Resource p, RDFNode o) throws ModelException {

    return m.find(s, p, o).size();
  }
  
  public Resource unique() throws ModelException {

    return m.getNodeFactory().createUniqueResource();
  }

  public boolean instanceOf(Resource i, Resource cls) throws ModelException {

    return RDFUtil.isInstanceOf(this, i, cls);
  }

  public void addAll(Model that) throws ModelException {

    SetOperations.unite(m, that);
  }

  /** iterates over all statements */
  public Iterator iterator() throws ModelException {

    return new IteratorST(elements());
  }

  /** iterates over statements */
  public Iterator iterator(Resource s, Resource p, RDFNode o) throws ModelException {

    return new IteratorST(m.find(s, p, o).elements());
  }

  /** iterates over objects of statements */
  public Iterator iteratorO(Resource s, Resource p) throws ModelException {

    return new IteratorO(m.find(s, p, null).elements());
  }

  class IteratorST implements Iterator {

    Enumeration en; // over statements
    Statement st;

    public IteratorST(Enumeration en) {

      this.en = en;
    }

    public boolean hasNext() {

      return en.hasMoreElements();
    }

    public Object next() {

      return st = (Statement)en.nextElement();
    }

    public void remove() {

      if(st != null) {
	try {
	  PowerModel.this.m.remove(st);
	} catch (ModelException any) {}
      }
    }
  }

  class IteratorO extends IteratorST {

    public IteratorO(Enumeration en) {

      super(en);
    }

    public Object next() {

      Statement st = (Statement)en.nextElement();
      try {
	return st.object();
      } catch (ModelException any) {
      }
      return null;
    }
  }
}
