/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package edu.stanford.db.rdf.vocabulary;

import org.w3c.rdf.vocabulary.rdf_syntax_19990222.RDF;
import org.w3c.rdf.vocabulary.rdf_schema_19990303.RDFS;
import org.w3c.rdf.vocabulary.dublin_core_19990702.DC;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.*;
import org.xml.sax.*;

import java.util.*;
import java.io.*;

/**
 * Generates Java classes with constants for resources defined in an (RDF) schema.
 * Java comments are generated from <tt>rdfs:comment</tt> and <tt>dc:Description</tt>.
 */
public class Generator {

  static final String DEFAULT_PACKAGE_CLASS = "UnspecifiedClass";

  static final String NS_IMPORT =
    "import org.w3c.rdf.model.*;\n" +
    "import java.util.*;\n";

  //  static final String DEFAULT_NODE_FACTORY = "new org.w3c.rdf.implementation.model.NodeFactoryImpl()";

  static final String DEFAULT_NODE_FACTORY = "new org.w3c.rdf.util.RDFFactoryImpl().getNodeFactory()";

  static final String NS_COMMENT = "/**\n" +
    " * This class provides convenient access to schema information.\n"+
    " * DO NOT MODIFY THIS FILE.\n" +
    " * It was generated automatically by edu.stanford.db.rdf.vocabulary.Generator\n" +
    " */\n";

  static final String NS_NSDEF = "  /** Namespace URI of this schema */";
  static final String NS_ID = "_Namespace";


  static Hashtable reservedWords = new Hashtable();

  static { // some obvious reserved words, FIXME: list incomplete
    addReservedWord("import");
    addReservedWord("class");
    addReservedWord("interface");
    addReservedWord("public");
    addReservedWord("private");
    addReservedWord("extends");
    addReservedWord("implements");
    addReservedWord("final");
    addReservedWord("int");
    addReservedWord("float");
    addReservedWord("double");
    addReservedWord("long");
    addReservedWord("boolean");
    addReservedWord("throws");
    addReservedWord("catch");
    addReservedWord("void");
    addReservedWord("default");
    addReservedWord("static");
    addReservedWord("transient");
    // namespace
    addReservedWord(NS_ID);
  }

  static void addReservedWord(String s) {
    reservedWords.put(s, s);
  }

  static boolean isReservedWord(String s) {
    return reservedWords.contains(s);
  }

  /**
   * Schema as input parameter
   */
  static void createJavaVocabulary(String packageClass, Model m, String namespace, String outputDirectory, String factoryStr) throws Exception {
                
    String packageName = null, className = null;
    if(factoryStr == null)
      factoryStr = DEFAULT_NODE_FACTORY;

    int ld = packageClass.lastIndexOf('.');
    if(ld < 0) {
      className = packageClass;
      packageName = "";
    } else {
      className = packageClass.substring(ld + 1);
      packageName = packageClass.substring(0, ld);
    }

    System.err.println("Creating interface " + className + " within package " + packageName +
                       (outputDirectory != null ? " in " + outputDirectory : ""));
                
    String packageDirectory = prepareOutputDir(packageName, outputDirectory);

    OutputStream out = openFileStream(packageDirectory, className, outputDirectory);
    dumpVocabulary( out, packageName, className, m, namespace, factoryStr );
    closeFileStream( out );
  }

  static String toJavaName(String s) {

    if(isReservedWord(s))
      return "_" + s;

    StringBuffer r = new StringBuffer();
    for(int i = 0; i < s.length(); i++) {
      char c = s.charAt(i);
      if(c == '-' || c == '.')
        r.append('_');
      else if(!Character.isJavaIdentifierPart(c))
        r.append("c" + (int)c);
      else
        r.append(c);
    }

    if(Character.isDigit(s.charAt(0)))
      return "_" + r;
    else
      return r.toString();
  }

  static void dumpVocabulary(OutputStream out, String packageName, String className, Model m, String ns, String factoryStr) throws Exception {
    
    PrintWriter p = new PrintWriter( new OutputStreamWriter(out) );

    if(!"".equals(packageName))
      p.println("package " + packageName + ";\n");
    p.println(NS_IMPORT);
    p.println(NS_COMMENT);
    p.println("public class " + className + " {\n");
    p.println(NS_NSDEF);
    p.println("  public static final String " + NS_ID + " = \"" + ns + "\";\n");
    p.println("  public static Set ALL;\n");

    Vector v = new Vector();

    // write resource declarations
    for(Enumeration en = RDFUtil.getResources(m).elements(); en.hasMoreElements(); ) {
      
      Resource r = (Resource)en.nextElement();
      String res = r.toString();
      //      String res = en.nextElement().toString();
      //      System.err.println("RESOURCE: " + res + ", ns=" + ns);
      if(res.startsWith(ns)) {
        String name = res.substring(ns.length());

        if(name.length() > 0) { // NS already included as a string

	  v.addElement(name);

          // comment?
          Statement tComment = RDFUtil.get1( m.find( r, RDFS.comment, null ) );
	  if(tComment == null)
	    tComment = RDFUtil.get1( m.find( r, DC.Description, null ) );
          if(tComment != null) {
            p.println("  /** " + tComment.object() + " */");
          }
          p.println("  public static Resource " + toJavaName(name) + ";\n");
        }
      }
    }

    p.println("  static {\n" +
	      "    try {\n" +
	      "      setNodeFactory(" + factoryStr + ");\n" +
	      "    } catch (ModelException ex) { ex.printStackTrace(System.err); }\n" +
	      "  }\n\n" +
	      "  public static " + className + " create() {\n" +
	      "    return new " + className + "();\n" +
	      "  }\n\n" +
	      "  private static Resource createResource(NodeFactory f, String suffix) throws ModelException {\n" +
	      "    Resource r = f.createResource(_Namespace, suffix);\n" +
	      "    ALL.add(r);\n" +
	      "    return r;\n" +
	      "  }\n\n" +
	      "  public static void setNodeFactory(NodeFactory f) throws ModelException {\n\n" +
	      "    ALL = new HashSet();\n");

    for(int i = 0; i < v.size(); i++) {

      String name = (String)v.elementAt(i);
      String jname = toJavaName(name);
      p.println("    " + jname + " = createResource(f, \"" + name + "\");");
    }

    p.println("  }\n}");
    p.flush();
  }

  static void closeFileStream(OutputStream out) throws IOException {
    
    if(out != System.out)
      out.close();
  }

  static OutputStream openFileStream(String packageDirectory, String className, String outputDirectory ) throws IOException {

    if(outputDirectory == null)
      return System.out;
                
    File f = new File( packageDirectory + File.separator + className + ".java");
    System.err.println("WRITING " + f);
    // replace existing file
    new RandomAccessFile(f, "rw").close();
    return new BufferedOutputStream( new FileOutputStream( f ) );
  }

  static String prepareOutputDir(String packageName, String outputDirectory) throws IOException {
    
    if(outputDirectory == null)
      return null;

    if(packageName == null)
      packageName = "";
    
    File dir = new File( outputDirectory );
    if(!dir.isDirectory())
      throw new IOException("Invalid output directory: " + outputDirectory);
    
    // follow package names
    for(StringTokenizer st = new StringTokenizer(packageName, "."); st.hasMoreTokens();) {
      dir = new File( dir, st.nextToken() );
      if(!dir.exists())
        dir.mkdir();
      else if (!dir.isDirectory())
        throw new IOException("Cannot create directory: " + dir);
    }
    return dir.getPath();
  }



  public static void main(String[] args) {
    
    RDFFactory f = new RDFFactoryImpl();
    Hashtable schemas = new Hashtable();

    String outputDirectory = null, packageClass = null, namespace = null, factoryStr = null;
      
    // parse options
    int par = 0;
    while (par < args.length) {

      if(args[par].startsWith("-d")) {
        outputDirectory = args[par+1];
        par += 2;
                                
      } else if(args[par].startsWith("-o")) {
        packageClass = args[par+1];
        par += 2;

      } else if(args[par].startsWith("-s")) {
        schemas.put(args[par+1], args[par+1]);
        par += 2;

      } else if(args[par].startsWith("-n")) {
        namespace = args[par+1];
        par += 2;

      } else if(args[par].startsWith("-f")) {
        factoryStr = args[par+1];
        par += 2;

      } else { // error
	schemas.clear();
        break;
      }
    }

    if(schemas.size() == 0) {
      System.err.println("Usage: Generator  [-d <output directory>] [-o <output class name>] [-n <namespace>] [-f <node factory expression>] {-s <schema file or URL>}+\n" +
                         "If -d is not specified, generated class will be printed to standard output.");
      System.exit(1);
    }

    if(packageClass == null)
      packageClass = DEFAULT_PACKAGE_CLASS;

    try {
      try {

	Model all = null;

	for(Enumeration en = schemas.elements(); en.hasMoreElements();) {

	  String schemaURL = (String)en.nextElement();
	  if(namespace == null)
	    namespace = schemaURL;

	  Model m = f.createModel();
	  RDFUtil.parse(schemaURL, f.createParser(), m);
	  System.err.println("Reading schema from " + schemaURL);

	  if(all == null)
	    all = m;
	  else
	    SetOperations.unite(all, m);
                                
	}

	System.err.println("Total statements: " + all.size());
	Generator.createJavaVocabulary(packageClass, all, namespace, outputDirectory, factoryStr);
      
      } catch (SAXException se) {
	System.err.println("Exception during parsing: " + se.getMessage());
	throw se.getException() != null ? se.getException() : se;
      }

    } catch (Exception any) {
      System.err.println("Could not generate vocabulary: " + any);
      any.printStackTrace(System.err);
    }
  }


}

